# -*- coding: utf-8 -*-
class BizException(Exception):
    def __init__(self, message: str, code: int = -1):
        self.message = message
        self.code = code


class NoAuthException(Exception):
    def __init__(self, message: str = '权限不足'):
        self.message = message
        self.code = 401


# 系统错误类异常
class SystemException(Exception):
    def __init__(self, message: str = '系统异常，请联系管理员处理'):
        self.message = message
        self.code = 500


# 系统参数为配置
class NoConfigException(SystemException):
    def __init__(self, message: str = '系统参数配置异常'):
        self.message = message
        self.code = 500




# -*- coding: utf-8 -*-
import json
import logging
from typing import List, Set, Optional, Dict
from ..client.redis_client import Redis
from ..lib import cfg
from ..model.items import AuthUser, AuthApp

logger = logging.getLogger(__name__)
rds = Redis(redis_uri=cfg.get('AUTH_REDIS_URL', None), redis_db=cfg.get('AUTH_REDIS_DB', None)).client
default_expire = 43200


def save_cached_app(access_token: str, app: AuthApp, expire: int = None):
    if app:
        _expire = expire or default_expire
        rds.set(f'share:access:{access_token}', app.json(), ex=_expire)


def get_cached_user(access_token: str) -> Optional[AuthUser]:
    rs = rds.get(f'share:access:{access_token}')
    if rs:
        js_user = json.loads(rs)
        # 容错
        if isinstance(js_user, str):
            js_user = json.loads(js_user)
        # 特征验证（user对象需要有id字段）
        if js_user and js_user.get('id', None):
            user = AuthUser(**js_user)
            user.userName = js_user.get('username', '')
            return user

    return None


def get_cached_app(access_token: str) -> Optional[AuthApp]:
    rs = rds.get(f'share:access:{access_token}')
    if rs:
        js_app = json.loads(rs)
        # 容错
        if isinstance(js_app, str):
            js_app = json.loads(js_app)
        # 特征验证（APP对象需要有appId字段）
        if js_app and js_app.get('appId', None):
            app = AuthApp(**js_app)
            return app

    return None


def get_cached_access_dict(access_token: str) -> Optional[Dict]:
    rs = rds.get(f'share:access:{access_token}')
    if rs:
        item = json.loads(rs)
        # 容错
        if isinstance(item, str):
            item = json.loads(item)
        return item

    return None


def has_access(access_token: str) -> bool:
    return rds.exists(f'share:access:{access_token}')


def save_cached_permission(access_token: str, permissions: Set[str], expire: int = None):
    if permissions:
        key = f'share:permissions:{access_token}'
        for per in permissions:
            rds.sadd(key, per)
        _expire = expire or default_expire
        rds.expire(key, _expire)


def has_permission(access_token: str, permission: str) -> bool:
    # 后者兼容Java版本异常
    _key = f'share:permissions:{access_token}'
    return rds.sismember(_key, f'{permission}') or rds.sismember(_key, f'"{permission}"')


def get_cached_permission(access_token: str) -> List[str]:
    return rds.smembers(f'share:permissions:{access_token}')


def clean_cache(access_token: str):
    rds.delete(f'share:access:{access_token}')
    rds.delete(f'share:permissions:{access_token}')



"""
Time Utilities

fast approaches to commonly used time/date related functions
"""
import datetime
import math
import time
from typing import Optional

import pytz
from dateutil import tz


def now() -> datetime.datetime:
    """
    get datetime instance of time of now
    :return: time of now
    """
    return datetime.datetime.now(tz.gettz('Asia/Shanghai'))


def __get_t(t: Optional[datetime.datetime] = None) -> datetime.datetime:
    """
    get datetime instance
    :param t: optional datetime instance
    :return: datetime instance
    """
    return t if isinstance(t, datetime.datetime) else now()


def to_str(t: Optional[datetime.datetime] = None,
           fmt: str = '%Y-%m-%d %H:%M:%S.%f') -> str:
    """
    get string formatted time
    :param t: optional datetime instance
    :param fmt: string format
    :return:
    """
    return __get_t(t).strftime(fmt)


def to_seconds(t: Optional[datetime.datetime] = None) -> int:
    """
    datetime to seconds
    :param t: optional datetime instance
    :return: timestamp in seconds
    """
    return int(__get_t(t).timestamp())


def to_milliseconds(t: Optional[datetime.datetime] = None) -> int:
    """
    datetime to milliseconds
    :param t: datetime instance
    :return: timestamp in seconds
    """
    return int(__get_t(t).timestamp() * 10 ** 3)


def to_microseconds(t: Optional[datetime.datetime] = None) -> int:
    """
    datetime to microseconds
    :param t: datetime instance
    :return: timestamp in seconds
    """
    return int(__get_t(t).timestamp() * 10 ** 6)


def get_dt(start_t: datetime.datetime,
           end_t: Optional[datetime.datetime] = None) -> datetime.timedelta:
    """
    get delta time
    :param start_t: start time
    :param end_t: end time
    :return: timedelta instance
    """
    return __get_t(end_t) - start_t


def to_seconds_dt(dt: datetime.timedelta) -> int:
    """
    delta time to seconds
    :param dt: timedelta instance
    :return: seconds elapsed
    """
    return int(dt.total_seconds())


def to_milliseconds_dt(dt: datetime.timedelta) -> int:
    """
    delta time to milliseconds
    :param dt: timedelta instance
    :return: milliseconds elapsed
    """
    return int(dt.total_seconds() * 10 ** 3)


def to_microseconds_dt(dt: datetime.timedelta) -> int:
    """
    delta time to microseconds
    :param dt: timedelta instance
    :return: microseconds elapsed
    """
    return int(dt.total_seconds() * 10 ** 6)


def parse_timestamp(time_str, tz_str='Asia/Shanghai'):
    """
    将时间戳解析成本地时间(自动截断毫秒)
    :param time_str:
    :param tz_str:
    :return:
    """
    if time_str:
        # 1576839034000
        if len(str(time_str)) > 10:
            # 截取掉毫秒
            time_str = str(time_str)[0:10]
        return datetime.datetime.fromtimestamp(int(time_str)).astimezone(tz.gettz(tz_str))


def time_to_batch_no(dt, delta_hour=0, delta_day=0, err=''):
    """
    # 将采购时间转换为批次编号(delta为0时，时间应为iso时间)，delta为正则加为负则减
    """
    try:
        if dt and isinstance(dt, datetime.datetime):
            if delta_hour or delta_day:
                time_delta = datetime.timedelta(hours=int(delta_hour), days=int(delta_day))
                dt = dt + time_delta
            return int(dt.strftime("%Y%m%d"))
    except ValueError:
        return err


def parse_time(time_str, fmt='%Y-%m-%d %H:%M:%S', tz_str='Asia/Shanghai'):
    """
    将时间字符串解析成本地时间
    :param time_str:
    :param fmt:
    :param tz_str:
    :return:
    """
    if time_str:
        return datetime.datetime.strptime(time_str, fmt).astimezone(tz.gettz(tz_str))


def parse_date(time_str, fmt='%Y-%m-%d', tz_str='Asia/Shanghai'):
    """
    将日期字符串解析成本地日期
    :param time_str:
    :param fmt:
    :param tz_str:
    :return:
    """
    if time_str:
        return datetime.datetime.strptime(time_str, fmt).astimezone(tz.gettz(tz_str))


def format_timestamp(time_str, fmt='%Y-%m-%d %H:%M:%S', tz_str='Asia/Shanghai'):
    """
    将时间戳解析成本地时间(自动截断毫秒)
    :param time_str:
    :param tz_str:
    :return:
    """
    if time_str:
        # 1576839034000
        if len(str(time_str)) > 10:
            # 截取掉毫秒
            time_str = str(time_str)[0:10]
        dt = datetime.datetime.fromtimestamp(int(time_str)).astimezone(tz.gettz(tz_str))
        return dt.strftime(fmt)


def format_time(dt, fmt='%Y-%m-%d %H:%M:%S'):
    """
    将时间格式化为字符串
    :param dt:
    :param fmt:
    :return:
    """
    if dt:
        return dt.strftime(fmt)


def format_date(dt, fmt='%Y-%m-%d'):
    """
    将时间格式化为字符串
    :param dt:
    :param fmt:
    :return:
    """
    if dt:
        return dt.strftime(fmt)


def get_timestamp13() -> int:
    """
    获取当前时间的时间戳 13位
    """
    # 生成13时间戳   eg:1540281250399895
    datetime_now = now()

    # 10位，时间点相当于从UNIX TIME的纪元时间开始的当年时间编号
    date_stamp = str(int(time.mktime(datetime_now.timetuple())))

    # 3位，微秒
    data_microsecond = str("%06d" % datetime_now.microsecond)[0:3]

    date_stamp = date_stamp + data_microsecond
    return int(date_stamp)


def add_time(start_time: Optional[datetime.datetime] = None, delta_hour: float = 0,
             delta_day: float = 0) -> datetime.datetime:
    """
    时间计算
    start_time: 起始时间 默认当前时间
    delta_hour: 偏移小时，正数为加，负数为减
    delta_day: 偏移天数，正数为加，负数为减
    """
    if not start_time:
        start_time = now()
    return start_time + datetime.timedelta(hours=delta_hour, days=delta_day)


def calc_range_seconds(start_time: datetime.datetime, end_time: datetime.datetime) -> int:
    """
    时间起止时间之间相差的总秒数
    start_time: 起始时间
    end_time: 结束时间
    """
    if start_time and end_time:
        total_seconds = (end_time - start_time).total_seconds()
        return int(total_seconds)


def calc_range_hour(start_time: datetime.datetime, end_time: datetime.datetime) -> float:
    """
    时间起止时间之间相差的总小时数
    start_time: 起始时间
    end_time: 结束时间
    """
    if start_time and end_time:
        total_seconds = (end_time - start_time).total_seconds()
        return float(int(total_seconds) / 3600)


def calc_range_day(start_time: datetime.datetime, end_time: datetime.datetime) -> int:
    """
    时间起止时间之间相差的总天数，不足一天的舍掉
    start_time: 起始时间
    end_time: 结束时间
    """
    if start_time and end_time:
        return (end_time - start_time).days


def current_timestamp13():
    """
    13位当前时间时间戳（毫秒级时间戳）
    :return:
    """
    return int(round(time.time() * 1000))


def current_timestamp10():
    """
    10位当前时间时间戳（秒级时间戳）
    :return:
    """
    return int(time.time())


def current_time_str(fmt: str = '%Y-%m-%d %H:%M:%S'):
    """
    10位当前时间时间戳（秒级时间戳）
    :return:
    """
    return format_time(now(), fmt)


def format_timestamp10(t: int, fmt: str = '%Y-%m-%d %H:%M:%S'):
    """
    10位当前时间时间戳（秒级时间戳）
    :return:
    """
    dt = parse_timestamp(t)
    return format_time(dt, fmt)


# 将时间转换成时间戳
def time_to_timestamp(time_str, fmt="%Y-%m-%d %H:%M:%S"):
    if time_str:
        timestamp = int(time.mktime((time.strptime(time_str, fmt))))

        return timestamp

    return ''


def get_time_stamp(result):
    '''
    tz时间格式字符串转换为时间戳
    :param result:
    :return:
    '''
    utct_date1 = datetime.datetime.strptime(result, "%Y-%m-%dT%H:%M:%S.%f%z")
    utct_date2 = time.strptime(result, "%Y-%m-%dT%H:%M:%S.%f%z")  # time.struct_time(tm_year=2020, tm_mon=12, tm_mday=1, tm_hour=3, tm_min=21, tm_sec=57, tm_wday=1, tm_yday=336, tm_isdst=-1)
    # print(utct_date1)
    # print(utct_date2)
    # 加上时区
    local_date = utct_date1 + datetime.timedelta(hours=8)
    local_date_srt = datetime.datetime.strftime(local_date, "%Y-%m-%d %H:%M:%S.%f")
    # print(local_date_srt)
    time_array1 = time.mktime(time.strptime(local_date_srt, "%Y-%m-%d %H:%M:%S.%f"))
    time_array2 = int(time.mktime(utct_date2))  # 1606764117
    # print(time_array1)
    # print(time_array2)
    return time_array2


# 获取当前时间到目标时间之间的秒数（用于做过期时间）
def get_second_from_now(end_time):
    if end_time and isinstance(end_time, datetime.datetime):
        if not end_time.tzinfo:
            end_time = end_time.astimezone(pytz.timezone('Asia/Shanghai'))
        total_seconds = (end_time - datetime.datetime.now(pytz.timezone('Asia/Shanghai'))).total_seconds()
        return math.floor(total_seconds)

    return 0


if __name__ == '__main__':
    print(time_to_timestamp('2023-06-27 00:00:00'))
    # print(get_time_stamp("2023-06-29T16:00:00.000Z"))
    # time_str = '2021-05-17'
    # print(time_to_timestamp(time_str))
    # print(parse_timestamp(str(int(now().timestamp()))))
    # print(str(1650591873000)[0:10])
    # print(time_to_timestamp("2023-06-29T16:00:00.000Z"))



# -*- coding: UTF-8 -*-
"""常用工具类"""
import math
import re
import random
from datetime import datetime
from dateutil.tz import tz


def half_split_int_range(start_num: int, end_num: int, must_int: bool = True):
    """
    将两个数字范围分割成等长的两个区间
    :param s:
    :param err:
    :return:
    """
    try:
        if start_num < end_num:
            half_num = start_num + (end_num - start_num) / 2
            # 向下取整转整数
            floor_half_num = math.floor(half_num)
            if floor_half_num <= start_num:
                if not must_int:
                    # [3, 4] ==> [(3, 3.5), (3.5, 4)]
                    return [(start_num, half_num), (half_num, end_num)]
                else:
                    return []
            else:
                return [(start_num, floor_half_num), (floor_half_num, end_num)]
        return []
    except ValueError:
        return []


def half_split_float_range(start_num: float, end_num: float):
    """
    将两个数字范围分割成等长的两个区间
    :param start_num:
    :param end_num:
    :return:
    """
    try:
        if start_num < end_num:
            half_num = start_num + (end_num - start_num) / 2
            return [(start_num, half_num), (half_num, end_num)]
        return []
    except ValueError:
        return []


def parse_number(s, err=0):
    """
    解析数字，如果解析失败返回默认值
    :param s:
    :param err:
    :return:
    """
    try:
        val = s
        if isinstance(s, str):
            s = s.replace('¥', '').replace('￥', '').replace(',', '').replace('元', '').strip()
            val = float(s)
        # 防止NaN值出现（如excel中）
        if math.isnan(val):
            return err
        return val
    except ValueError:
        return err
    except Exception:
        return err


def parse_int(s, err=0):
    """
    解析整型数字，如果解析失败返回默认值
    :param s:
    :param err:
    :return:
    """
    try:
        if isinstance(s, str):
            s = s.replace('¥', '').replace('￥', '').replace(',', '').replace('元', '').strip()
        return int(s)
    except ValueError:
        return err
    except TypeError:
        return err


def format_timestamp(time_str, fmt='%Y-%m-%d %H:%M:%S', tz_str='Asia/Shanghai'):
    """
    将时间戳解析成本地时间(自动截断毫秒)
    :param time_str:
    :param tz_str:
    :return:
    """
    if time_str:
        # 1576839034000
        if len(str(time_str)) > 10:
            # 截取掉毫秒
            time_str = str(time_str)[:-3]
        dt = datetime.fromtimestamp(int(time_str)).astimezone(tz.gettz(tz_str))
        return dt.strftime(fmt)


def join_dict(kv):
    """
    将dict键值对拼接成字符串
    :param kv:
    :return:
    """
    s = ''
    for (k, v) in kv.items():
        s += k + ': ' + v + '\r\n'
    return s.strip()


def random_code(n=5, alpha=True):
    """
    生成验证码
    :param n:
    :param alpha:
    :return:
    """
    s = ''  # 创建字符串变量,存储生成的验证码
    for i in range(n):  # 通过for循环控制验证码位数
        num = random.randint(0, 9)  # 生成随机数字0-9
        if alpha:  # 需要字母验证码,不用传参,如果不需要字母的,关键字alpha=False
            upper_alpha = chr(random.randint(65, 90))
            lower_alpha = chr(random.randint(97, 122))
            num = random.choice([num, upper_alpha, lower_alpha])
        s = s + str(num)
    return s


def match_ip_port(txt):
    """
    解析IP和端口
    :param txt:
    :return:
    """
    # /frontBrands/getBrandsAndProductInfos.action?orderBy=normal&shopInfoId=1
    if txt:
        arr = re.findall(r'(?:(?:[0,1]?\d?\d|2[0-4]\d|25[0-5])\.){3}(?:[0,1]?\d?\d|2[0-4]\d|25[0-5]):\d{0,5}',
                         txt.strip())
        if len(arr):
            return arr[0].strip()
    return ''


def list_of_groups(init_list, childern_list_len):
    """
    列表分割成多个小列表
    :param init_list:
    :param childern_list_len:
    :return:
    """
    list_of_groups = zip(*(iter(init_list),) * childern_list_len)
    end_list = [list(i) for i in list_of_groups]
    count = len(init_list) % childern_list_len
    end_list.append(init_list[-count:]) if count != 0 else end_list
    return end_list


if __name__ == '__main__':
    # print(half_split_float_range(5, 8))
    # print(half_split_int_range(3, 6))
    # print(half_split_int_range(3, 5))
    # print(half_split_int_range(3, 4))
    # print(get_timestamp13())
    # print(format_timestamp(int(time.time() * 1000), fmt="%Y/%m/%d"))
    pass



# -*- coding: utf-8 -*-
import os
import typing
from datetime import timedelta
from typing import Optional, Dict
from minio import Minio
from minio.error import NoSuchKey
from ..exception.exceptions import NoConfigException
from ..lib import cfg, logger

LOGGER = logger.get('MinIOClient')


# MinIO客户端简易封装
class MinIO(object):

    def __init__(self,
                 endpoint: str = None,
                 access_key: str = None,
                 secret_key: str = None,
                 region: str = None,
                 tmp_dir: str = None):
        self.endpoint_private = endpoint or cfg.get('MINIO_ENDPOINT_PRIVATE') or cfg.get('MINIO_ENDPOINT_PUBLIC')
        self.endpoint_public = cfg.get('MINIO_ENDPOINT_PUBLIC')
        self.endpoint_public_scheme = cfg.get('MINIO_ENDPOINT_PUBLIC_SCHEME', 'http')
        if not self.endpoint_private:
            raise NoConfigException('MinIO endpoint not config!')
        self.access_key = access_key or cfg.get('MINIO_ACCESS_KEY')
        if not self.access_key:
            raise NoConfigException('MinIO access_key not config!')
        _secret_key = secret_key or cfg.get('MINIO_SECRET_KEY')
        if not _secret_key:
            raise NoConfigException('MinIO secret_key not config!')
        self.region = region or cfg.get('MINIO_REGION') or 'region-default'
        if not self.region:
            raise NoConfigException('MinIO region not config!')
        _secure = cfg.get_bool('MINIO_SECURE') or False
        self.max_presigned_day = cfg.get_int('MINIO_PRESIGNED_DAY') or 7
        self.tmp_dir = tmp_dir or cfg.get_bool('MINIO_TMP_DIR') or '/tmp'
        if not self.tmp_dir:
            raise NoConfigException('MinIO tmp_dir not config!')

        self.client = Minio(
            endpoint=self.endpoint_private,
            access_key=self.access_key,
            secret_key=_secret_key,
            region=self.region,
            secure=_secure
        )

        self.public_client = Minio(
            endpoint=self.endpoint_public,
            access_key=self.access_key,
            secret_key=_secret_key,
            region=self.region,
            secure=_secure
        )

    def put_stream(self, bucket_name: str, object_name: str, image_content: bytes, content_type='application/octet-stream') -> Dict:
        metadata = dict()
        metadata['Content-Type'] = content_type or 'application/octet-stream'
        self.client._do_put_object(bucket_name, object_name, image_content, len(image_content), metadata=metadata)
        return {
            'url': f'{self.endpoint_public_scheme}://{self.endpoint_public}/{bucket_name}/{object_name}',
            'file_path': f'{bucket_name}/{object_name}',
        }

    def put_object(self, bucket_name: str, object_name: str, file: typing.IO, content_type='application/octet-stream') -> Dict:
        self.client.put_object(bucket_name, object_name, file, os.fstat(file.fileno()).st_size, content_type=content_type)

        return {
            'url': f'{self.endpoint_public_scheme}://{self.endpoint_public}/{bucket_name}/{object_name}',
            'file_path': f'{bucket_name}/{object_name}',
        }

    def get_download_url(self, bucket_name: str, object_name: str) -> Optional[str]:
        try:
            result = self.client.stat_object(bucket_name, object_name)
            if result and result.last_modified and result.size:
                url = self.public_client.presigned_get_object(
                    bucket_name=bucket_name,
                    object_name=object_name,
                    expires=timedelta(days=self.max_presigned_day),
                    response_headers={"response-content-type": "application/zip"}
                )
                return url
        except NoSuchKey:
            LOGGER.info(f'包不存在: bucket_name={bucket_name}, object_name={object_name}')
        return None



import copy
from zcbot_web_core.exception.exceptions import SystemException
from zcbot_web_core.lib import logger
from zcbot_web_core.client.mongo_client import Mongo


class DynamicSettings(object):
    """
    简易配置管理器
    """
    # 主键索引
    row_map = {}
    # 组索引
    group_map = {}

    def __init__(self, collection: str, primary_field: str, group_field: str = None, value_field: str = None, mongo: Mongo = None, name: str = None, entity_class = None):
        if not collection:
            raise SystemException(f'[collection]集合未指定: id={id}, collection={self.collection}, primary_field={self.primary_field}, group_field={self.group_field}, value_field={self.value_field}')

        self.logger = logger.get(name or '配置')
        self.collection = collection
        self.primary_field = primary_field or '_id'
        self.value_field = value_field
        self.group_field = group_field
        self.entity_class = entity_class
        self.mongo = mongo or Mongo()
        self.reload()

    def get_value(self, key):
        if not self.value_field:
            self.logger.error(f'[value_field]字段未指定: collection={self.collection}, primary_field={self.primary_field}, group_field={self.group_field}, value_field={self.value_field}')
            return None

        row = self.row_map.get(key)
        if row:
            val = row.get(self.value_field, None)
            return copy.deepcopy(val)

        return None

    def get_row(self, key):
        row = self.row_map.get(key, None)

        return copy.deepcopy(row)

    def get_all(self):
        rows = [*self.row_map.values()]

        return copy.deepcopy(rows)

    def get_rows_by_group(self, group_key):
        if not self.group_field:
            self.logger.error(f'[group_field]字段未指定: collection={self.collection}, primary_field={self.primary_field}, group_field={self.group_field}, value_field={self.value_field}')
            return None

        rows = self.group_map.get(group_key, None)

        return copy.deepcopy(rows)

    def reload(self):
        items = self._fetch_items()
        hot_row_map = {}
        hot_group_map = {}
        for item in items:
            item_obj = item
            if self.entity_class:
                item_obj = self.entity_class(**item)
            id = item.get(self.primary_field, None)
            # 增加禁用功能（可选）
            if 'enable' in item and item.get('enable', None) != 1:
                self.logger.warning(f'禁用配置: id={id}, collection={self.collection}, primary_field={self.primary_field}, value_field={self.value_field}')
                continue

            # 主键索引
            if id:
                hot_row_map[id] = item_obj
            else:
                self.logger.warning(f'异常配置: id={id}, collection={self.collection}, primary_field={self.primary_field}, value_field={self.value_field}')

            # 分组索引
            if self.group_field:
                group_key = item.get(self.group_field, None)
                if group_key:
                    if group_key not in hot_group_map:
                        hot_group_map[group_key] = list()
                    rows = hot_group_map.get(group_key)
                    rows.append(item_obj)

        self.row_map = hot_row_map
        self.group_map = hot_group_map

        self.logger.info(f'更新配置: row={len(self.row_map.keys())}条 group={len(self.group_map.keys())}条, collection={self.collection}, primary_field={self.primary_field}, value_field={self.value_field}, group_field={self.group_field}')

    def _fetch_items(self):
        try:
            rs = self.mongo.list(collection=self.collection, query={})
            return rs
        except Exception as e:
            self.logger.error(e)

        return []



from pydantic import BaseModel, Field
from ..model.enums import UserStatus, UserGender
from ..lib import time as time_lib


class BaseData(BaseModel):
    """
    通用基础数据模型
    """
    # mongodb主键
    _id: str = None
    # 插入时间
    genTime: int = Field(
        default_factory=time_lib.current_timestamp10
    )


class AuthUser(BaseData):
    """
    用户鉴权模型
    """
    # 用户编号
    id: str
    # 租户编码
    tenantCode: str = None
    # 用户名
    userName: str = None
    # 用户姓名
    realName: str = None
    # 性别
    gender: int = UserGender.MALE
    # 电话
    mobile: str = None
    # 邮箱
    email: str = None
    # 是否超级管理员
    superAdmin: str = None
    # 是否超级租户
    superTenant: int = 1
    # 启用停用
    isEnabled: bool = True
    # 批次状态
    status: str = UserStatus.NORMAL

    def to_log(self):
        return f'{self.realName}({self.id})'

    def is_super_admin(self):
        return '1' == self.superAdmin


class AuthApp(BaseData):
    """
    App鉴权模型
    """
    # 用户编号
    appId: str
    # 用户名
    appName: str = None
    # 账号
    accessKey: str = None
    # 密码
    secretKey: str = None
    # 长效Token
    staticToken: str = None
    # 是否启用长效Token
    enableStaticToken: str = None
    # 启用停用
    isEnabled: bool = None
    # 批次状态
    status: str = None
    # 租户编码
    tenantCode: str = None



# from starlette.requests import Request
# from starlette.responses import Response
# from starlette.middleware.base import BaseHTTPMiddleware
# from ..auth.share_auth import ShareAuth
# from ..context.context_vars import auth_user_context, tenant_context
# from ..lib import logger
#
# LOGGER = logger.get('租户中间件')
# share_auth = ShareAuth()
#
#
# class TenantMiddleware(BaseHTTPMiddleware):
#     """ 租户中间件 """
#
#     def __init__(self, app):
#         super().__init__(app)
#
#     async def dispatch(self, request: Request, next_func) -> Response:
#         """
#         租户挂载中间件
#         :param request:
#         :param next_func:
#         :return:
#         """
#         # 挂载
#         auth_user = share_auth.get_auth_user(request)
#         auth_user_context.set(auth_user)
#         if auth_user and auth_user.tenantCode:
#             # 复用auth_user避免重复解析用户
#             tenant_context.set(auth_user.tenantCode)
#         # 执行
#         response = await next_func(request)
#         # 清理
#         tenant_context.set(None)
#         auth_user_context.set(None)
#
#         return response



from typing import List
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware import Middleware


def register_middleware(app: FastAPI, middlewares: List[Middleware] = None):
    app.add_middleware(
        CORSMiddleware,
        allow_credentials=True,
        allow_origins=['*'],
        allow_methods=["*"],
        allow_headers=["*"],
    )

    if middlewares:
        for middleware in middlewares:
            # if app.middleware_stack is not None:  # pragma: no cover
            #     raise RuntimeError("Cannot add middleware after an application has started")
            app.user_middleware.insert(0, middleware)



import pytz
import traceback
from pymongo import MongoClient
from apscheduler.jobstores.mongodb import MongoDBJobStore
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.jobstores.base import JobLookupError, ConflictingIdError
from ..exception.exceptions import BizException
from ..lib import logger, cfg

LOGGER = logger.for_job('scheduler_server')

scheduler = AsyncIOScheduler()


def init_scheduler():
    global scheduler

    try:
        # enable_scheduler = cfg.get_bool('ENABLE_SCHEDULER')
        # if not enable_scheduler:
        #     LOGGER.info('定时任务框架未启用')
        #     return
        tzinfo = pytz.timezone('Asia/Shanghai')
        default_mongo_jobstore = MongoDBJobStore(
            client=MongoClient(cfg.get('MONGO_URL'), tz_aware=True, tzinfo=tzinfo),
            database=cfg.get('MONGO_DB'),
            collection=cfg.get('SCHEDULER_JOB_STORE_NAME', 'sys_jobs')
        )
        job_stores = {
            'default': default_mongo_jobstore
        }
        executors = {
            'default': {'type': 'threadpool', 'max_workers': 1000},
        }

        scheduler.configure(
            jobstores=job_stores,
            executors=executors,
        )
        scheduler.start()
        LOGGER.info('定时任务框架初始化成功')

    except Exception as e:
        LOGGER.error(f'定时任务框架初始化异常： {traceback.format_exc()}')


def _env_check():
    # enable_scheduler = cfg.get_bool('ENABLE_SCHEDULER')
    # if not enable_scheduler:
    #     return False
    if not scheduler:
        raise BizException('定时任务功能未开启或框架初始化失败')
    return True


def get_scheduler():
    if _env_check():
        return scheduler


def add_fix_date_job(job_id, func, replace_existing=True, **options):
    """
    特定时间定时调度（作业只会执行一次）
    https://www.cnblogs.com/luxiaojun/p/6567132.html
    :param job_id: 定时任务编号
    :param func: 执行的函数
    :param options: 可选参数
        run_date (datetime|str) – the date/time to run the job at  -（任务开始的时间）
        timezone (datetime.tzinfo|str) – time zone for run_date if it doesn’t have one already
    :return:
    """
    try:
        if _env_check():
            LOGGER.info(f'add fix date job: job_id={job_id}, func={func}')
            scheduler.add_job(func, 'date', id=job_id, replace_existing=replace_existing, **options)
    except ConflictingIdError as err:
        raise BizException('任务已存在')


def add_interval_job(job_id, func, replace_existing=True, **options):
    """
    间隔调度任务（每隔多久执行）
    https://www.cnblogs.com/luxiaojun/p/6567132.html
    :param job_id: 定时任务编号
    :param func: 执行的函数
    :param options: 可选参数
        weeks (int) – number of weeks to wait
        days (int) – number of days to wait
        hours (int) – number of hours to wait
        minutes (int) – number of minutes to wait
        seconds (int) – number of seconds to wait
        start_date (datetime|str) – starting point for the interval calculation
        end_date (datetime|str) – latest possible date/time to trigger on
        timezone (datetime.tzinfo|str) – time zone to use for the date/time calculations
    :return:
    """
    try:
        if _env_check():
            LOGGER.info(f'add interval job: job_id={job_id}, func={func}')
            scheduler.add_job(func, 'interval', id=job_id, replace_existing=replace_existing, **options)
    except ConflictingIdError as err:
        raise BizException('任务已存在')


def add_cron_job(job_id, func, replace_existing=True, **options):
    """
    cron定时调度（某一定时时刻执行）
    https://www.cnblogs.com/luxiaojun/p/6567132.html
    :param job_id: 定时任务编号
    :param func: 执行的函数
    :param options: 任务时间配置可选参数
        year (int|str) – 4-digit year -（表示四位数的年份，如2008年）
        month (int|str) – month (1-12) -（表示取值范围为1-12月）
        day (int|str) – day of the (1-31) -（表示取值范围为1-31日）
        week (int|str) – ISO week (1-53) -（格里历2006年12月31日可以写成2006年-W52-7（扩展形式）或2006W527（紧凑形式））
        day_of_week (int|str) – number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun) - （表示一周中的第几天，既可以用0-6表示也可以用其英语缩写表示）
        hour (int|str) – hour (0-23) - （表示取值范围为0-23时）
        minute (int|str) – minute (0-59) - （表示取值范围为0-59分）
        second (int|str) – second (0-59) - （表示取值范围为0-59秒）
        start_date (datetime|str) – earliest possible date/time to trigger on (inclusive) - （表示开始时间）
        end_date (datetime|str) – latest possible date/time to trigger on (inclusive) - （表示结束时间）
        timezone (datetime.tzinfo|str) – time zone to use for the date/time calculations (defaults to scheduler timezone) -（表示时区取值）
    :return:
    """
    try:
        if _env_check():
            LOGGER.info(f'add cron job: job_id={job_id}, func={func}')
            scheduler.add_job(func, 'cron', id=job_id, replace_existing=replace_existing, **options)
    except ConflictingIdError as err:
        raise BizException('任务已存在')


def remove_job(job_id, ignore_error=True):
    try:
        if _env_check():
            LOGGER.info(f'remove job: job_id={job_id}, ignore_error={ignore_error}')
            scheduler.remove_job(job_id=job_id)
    except JobLookupError as err:
        if not ignore_error:
            raise BizException('任务不存在或已删除')
        else:
            LOGGER.error(f'任务不存在或已删除: job_id={job_id}')



# -*- coding: utf-8 -*-
import pymysql
from pymysql.cursors import DictCursor
from ..lib import cfg, logger

LOGGER = logger.get('MySQLClient')


class MySQL(object):
    """
    MySQL客户端简易封装
    """

    def __init__(self, mysql_host=None, mysql_port=None, mysql_user=None, mysql_password=None, mysql_db=None):
        self._biz_inited = True
        self.host = mysql_host or cfg.get('MYSQL_HOST')
        self.port = mysql_port or cfg.get('MYSQL_PORT')
        self.user = mysql_user or cfg.get('MYSQL_USER')
        self.pwd = mysql_password or cfg.get('MYSQL_PASSWORD')
        self.db = mysql_db or cfg.get('MYSQL_DB')
        self.connection = pymysql.connect(host=self.host, port=int(self.port), user=self.user, password=self.pwd, db=self.db)

    # 查询对象
    def get(self, sql, cursor_type=DictCursor):
        try:
            if self.connection and sql:
                with self.connection.cursor(cursor_type) as cursor:
                    cursor.execute(sql)
                    rs = cursor.fetchone()
                    return rs
            return {}
        except Exception as e:
            print(e)

    # 查询列表
    def list(self, sql, cursor_type=DictCursor):
        try:
            if self.connection and sql:
                with self.connection.cursor(cursor_type) as cursor:
                    cursor.execute(sql)
                    rs_list = cursor.fetchall()
                    return rs_list
            return list()
        except Exception as e:
            print(e)

    # 销毁关闭链接
    def __del__(self):
        if self.connection:
            try:
                self.connection.close()
            except Exception as e:
                print(e)



"""
Common/Misc utils
"""

import platform
import pprint
import typing
from typing import Any, Tuple, Optional

T = typing.TypeVar('T')


# global encoding
ENCODING = 'utf-8'


def as_int(o: Any) -> Tuple[int, Optional[Exception]]:
    """
    object to int
    :param o: object
    :return: int value, err
    """
    try:
        return int(o), None
    except Exception as e:
        return 0, e


def as_float(o: Any) -> Tuple[float, Optional[Exception]]:
    """
    object to float
    :param o: object
    :return: float value, err
    """
    try:
        return float(o), None
    except Exception as e:
        return 0.0, e


def is_win() -> bool:
    """
    is currently running on windows
    :return: if current system is of Windows family
    """
    return platform.system().lower().startswith('win')


def pfmt(o: Any, *args, **kwargs) -> str:
    """
    pretty format object
    :param o: object
    :return: object string
    """
    return pprint.pformat(o, *args, **kwargs)



# -*- coding: utf-8 -*-
import threading
from ..client.mongo_client import Mongo
from ..lib import cfg, logger

LOGGER = logger.get('MultiDBMongo')


class MultiDBMongo(Mongo):
    """
    多库Mongo客户端简易封装
    """

    def __init__(self, mongo_url=None, mongo_db=None):
        if not mongo_url:
            mongo_url = cfg.get('MONGO_URL') or 'mongodb://localhost:27017'
        if not mongo_db:
            mongo_db = cfg.get('MONGO_DB')
            if not mongo_db:
                raise Exception('mongodb database not specified')

        super().__init__(mongo_url, mongo_db)
        LOGGER.info(f'[{mongo_db}] multi database mongodb inited~')


class MultiDBMongoHolder(object):
    """
    多库Mongo客户端持有器（租户数据库连接池）
    """
    _instance_lock = threading.Lock()
    _db_instance_dict = dict()

    def __new__(cls, *args, **kwargs):
        if not hasattr(MultiDBMongoHolder, "_instance"):
            with MultiDBMongoHolder._instance_lock:
                if not hasattr(MultiDBMongoHolder, "_instance"):
                    MultiDBMongoHolder._instance = object.__new__(cls)
        return MultiDBMongoHolder._instance

    @staticmethod
    def get_mongo(db_name) -> MultiDBMongo:
        if not db_name:
            return None

        # 有实例则直接返回
        if db_name in MultiDBMongoHolder._db_instance_dict:
            return MultiDBMongoHolder._db_instance_dict.get(db_name)

        # 无实例则加锁创建
        MultiDBMongoHolder._instance_lock.acquire()
        try:
            # 双重锁校验
            if db_name not in MultiDBMongoHolder._db_instance_dict:
                # 初始化新实例
                inst = MultiDBMongo(mongo_db=db_name)
                MultiDBMongoHolder._db_instance_dict[db_name] = inst
            return MultiDBMongoHolder._db_instance_dict.get(db_name)
        finally:
            MultiDBMongoHolder._instance_lock.release()



# -*- coding: utf-8 -*-
import contextvars

# 租户环境变量
tenant_context = contextvars.ContextVar('tenant')

# # 当前登录用户环境变量
# auth_user_context = contextvars.ContextVar('auth_user')

# 非主线程的事件循环环境变量
event_loop_context = contextvars.ContextVar('event_loop')



from distutils.core import setup
from setuptools import find_packages

with open('README.rst', 'r') as f:
    long_description = f.read()

setup(name='zcbot-web-core',
      version='1.35',
      description='zcbot web core for zsodata',
      long_description=long_description,
      author='zsodata',
      author_email='team@zso.io',
      url='http://www.zsodata.com',
      install_requires=[
            # "fastapi==0.109.2",
            # "starlette==0.36.3",
            # "uvicorn==0.27.1",
            # "click==8.1.7",
            # "h11==0.14.0",
            # "pydantic==1.8.2",
            # "python-dotenv==0.19.0",
            # "python-multipart==0.0.5",
            # "six==1.16.0",
            # "tzlocal==2.1",
            # "pandas==1.0.5",
            # "httpx==0.26.0",
            # "httpx-socks==0.8.1",
            # "requests==2.31.0",
            # "numpy==1.22.3",
            # "pymongo==3.11.0",
            # "redis==3.5.3",
            # "python-dateutil==2.8.2",
            # "pytz==2024.1",
            # "tenacity==8.2.3",
            # "cryptography==42.0.4",
            # "APScheduler==3.10.4",
            # "oss2==2.12.1",
      ],
      python_requires='>=3.7',
      license='BSD License',
      packages=find_packages(),
      platforms=['all'],
      include_package_data=True
      )



# -*- coding: utf-8 -*-
import threading
from ..client.mongo_client import Mongo
from ..lib import cfg, logger

LOGGER = logger.get('MultiTenantMongo')


class MultiTenantMongo(Mongo):
    """
    多租户Mongo客户端简易封装
    """

    def __init__(self, tenant_code=None, mongo_url=None, mongo_db=None, db_prefix=None):
        if not mongo_url:
            mongo_url = cfg.get('MONGO_URL') or 'mongodb://localhost:27017'
        if not mongo_db:
            if tenant_code and (db_prefix or cfg.get('MONGO_DB_BIZ_PREFIX', None)):
                mongo_db = f"{db_prefix or cfg.get('MONGO_DB_BIZ_PREFIX', None)}{tenant_code}"
        if not mongo_db:
            raise Exception('mongodb database not specified')

        super().__init__(mongo_url, mongo_db)
        LOGGER.info(f'[{tenant_code}] tenant mongodb inited~')


class MultiTenantMongoHolder(object):
    """
    多租户Mongo客户端持有器（租户数据库连接池）
    """
    _instance_lock = threading.Lock()
    _tenant_instance_dict = dict()

    def __new__(cls, *args, **kwargs):
        if not hasattr(MultiTenantMongoHolder, "_instance"):
            with MultiTenantMongoHolder._instance_lock:
                if not hasattr(MultiTenantMongoHolder, "_instance"):
                    MultiTenantMongoHolder._instance = object.__new__(cls)
        return MultiTenantMongoHolder._instance

    @staticmethod
    def get_tenant_mongo(tenant_code) -> MultiTenantMongo:
        if not tenant_code:
            return None

        # 有实例则直接返回
        if tenant_code in MultiTenantMongoHolder._tenant_instance_dict:
            return MultiTenantMongoHolder._tenant_instance_dict.get(tenant_code)

        # 无实例则加锁创建
        MultiTenantMongoHolder._instance_lock.acquire()
        try:
            # 双重锁校验
            if tenant_code not in MultiTenantMongoHolder._tenant_instance_dict:
                # 初始化新实例
                inst = MultiTenantMongo(tenant_code=tenant_code)
                MultiTenantMongoHolder._tenant_instance_dict[tenant_code] = inst
            return MultiTenantMongoHolder._tenant_instance_dict.get(tenant_code)
        finally:
            MultiTenantMongoHolder._instance_lock.release()



from __future__ import absolute_import


def init_core_modules(*args, **kwargs) -> None:
    """
    Init core handlers, middlewares and services for fastapi application
    :return: None
    """
    pass



# -*- coding: utf-8 -*-
import logging
import re
import json
import threading
from typing import Optional, List, Union
from starlette.routing import BaseRoute
from fastapi.routing import APIRoute
from starlette.requests import Request
from ..lib import cfg
from ..auth import auth_cache_pool
from ..client.redis_client import Redis
from ..exception.exceptions import NoAuthException
from ..model.items import AuthUser, AuthApp

logger = logging.getLogger(__name__)

FIX_WHITELIST = [
]


class ShareAuth(object):
    _instance_lock = threading.Lock()
    _biz_inited = False
    _routes_inited = False

    def __new__(cls, *args, **kwargs):
        if not hasattr(ShareAuth, "_instance"):
            with ShareAuth._instance_lock:
                if not hasattr(ShareAuth, "_instance"):
                    ShareAuth._instance = object.__new__(cls)
        return ShareAuth._instance

    def __init__(self):
        if not self._biz_inited:
            self._biz_inited = True
            self.rds = Redis(redis_uri=cfg.get('AUTH_REDIS_URL', None), redis_db=cfg.get('AUTH_REDIS_DB', None)).client
            # route path regex -> tags
            self.regex_to_tags_map = dict()
            # no auth path whitelist
            self.auth_whitelist = []

    def get_access_token(self, request: Request):
        """
        解析请求头鉴权信息
        :param request:
        :return:
        """
        if request and (request.headers or request.query_params):
            access_token = request.headers.get('Authorization', '').replace('Bearer ', '')
            if not access_token:
                access_token = request.cookies.get('access_token')
            if not access_token:
                access_token = request.query_params.get('access_token')
            return access_token

        return ''

    def reload(self, routes: Optional[List[BaseRoute]], white_list=[], force: bool = False):
        if not force and self._routes_inited:
            return self
        # 构造路由匹配正则与权限颗粒(tags)映射
        _tmp_regex_to_tags_map = dict()
        for route in routes:
            if isinstance(route, APIRoute) and route.path_regex:
                _tmp_regex_to_tags_map[route.path_regex] = route.tags
        self.regex_to_tags_map.clear()
        self.regex_to_tags_map = _tmp_regex_to_tags_map
        logger.info(f'reload path to tags map: {len(self.regex_to_tags_map)}')

        # 构造权限白名单
        _auth_whitelist = list()
        for row in FIX_WHITELIST:
            _auth_whitelist.append(re.compile(row))
        for row in white_list:
            _auth_whitelist.append(re.compile(row))
        self.auth_whitelist.clear()
        self.auth_whitelist = _auth_whitelist
        logger.info(f'reload no auth whitelist: {len(self.auth_whitelist)}')
        #
        self._routes_inited = True

        return self

    def access_check(self, request: Request):
        """
        登录校验
        :param request:
        :return:
        """
        access_token = self.get_access_token(request)
        if not access_token:
            raise NoAuthException('请先登录')

        if not auth_cache_pool.has_access(access_token):
            raise NoAuthException('登录信息已失效')

        return access_token

    def auth_check(self, request: Request):
        """
        权限校验
        :param request:
        :return:
        """
        if not self.regex_to_tags_map:
            logger.debug(f'no path to tags map')
            return False
        # 优先处理白名单
        for white_regex in self.auth_whitelist:
            if white_regex.match(request.url.path):
                return True
        # 先校验登录
        access_token = self.access_check(request)
        for regex in self.regex_to_tags_map.keys():
            if regex.match(request.url.path):
                # 匹配到路由
                tags = self.regex_to_tags_map.get(regex)
                if tags:
                    # 需要登录且需要鉴权
                    auth_rs = self._rds_permissions_check(tags, access_token)
                    if not auth_rs:
                        raise NoAuthException('权限不足')
                    return True
                else:
                    # 仅登录无需鉴权
                    return True

        return False

    def _rds_permissions_check(self, tags, access_token) -> bool:
        # redis权限查询
        for tag in tags:
            # 新权限颗粒 xx.xx.xx.xx
            has = auth_cache_pool.has_permission(access_token, tag)
            if not has:
                # 兼容Java版权限颗粒 "xx.xx.xx.xx"
                has = auth_cache_pool.has_permission(access_token, json.dumps(tag))
                if not has:
                    return False

        return True

    def get_auth_user(self, request: Request) -> Optional[AuthUser]:
        """
        获取当前登录的用户信息
        :param request:
        :return:
        """
        access_token = self.get_access_token(request)
        if not access_token:
            raise NoAuthException('请先登录')
        if not auth_cache_pool.has_access(access_token):
            raise NoAuthException('登录信息已失效')

        auth_user = auth_cache_pool.get_cached_user(access_token)
        # 仅超级管理员可以切换租户
        if auth_user and auth_user.superAdmin == '1':
            tenant_code = request.headers.get('TenantCode', None)
            if tenant_code:
                # 切换租户
                auth_user.tenantCode = tenant_code

        return auth_user

    def get_auth_app(self, request: Request) -> Optional[AuthApp]:
        """
        获取当前登录的应用信息
        """
        access_token = self.get_access_token(request)
        if not access_token:
            raise NoAuthException('请先鉴权认证')
        if not auth_cache_pool.has_access(access_token):
            raise NoAuthException('鉴权认证信息已失效')

        return auth_cache_pool.get_cached_app(access_token=access_token)

    def get_auth_tenant_code(self, request: Request) -> Optional[str]:
        """
        获取授权用户所属租户
        :param request:
        :return:
        """
        access_token = self.get_access_token(request)
        if access_token:
            access_item = auth_cache_pool.get_cached_access_dict(access_token)
            if access_item:
                return access_item.get('tenantCode', '')

        return ''


# 用于handler来Depends，获取授权用户
async def authed_user(request: Request) -> Optional[AuthUser]:
    return ShareAuth().get_auth_user(request)


# 用于handler来Depends，获取授权APP信息
async def authed_app(request: Request) -> Optional[AuthApp]:
    return ShareAuth().get_auth_app(request)


# 用于handler来Depends，获取授权用户
async def access_token(request: Request) -> Optional[str]:
    return ShareAuth().get_access_token(request)



import six
from functools import partial
from inspect import isclass, ismethod


def get_callable_name(func):
    """
    Returns the best available display name for the given function/callable.
    :rtype: str
    """
    # the easy case (on Python 3.3+)
    if hasattr(func, '__qualname__'):
        return func.__qualname__

    # class methods, bound and unbound methods
    f_self = getattr(func, '__self__', None) or getattr(func, 'im_self', None)
    if f_self and hasattr(func, '__name__'):
        f_class = f_self if isclass(f_self) else f_self.__class__
    else:
        f_class = getattr(func, 'im_class', None)

    if f_class and hasattr(func, '__name__'):
        return '%s.%s' % (f_class.__name__, func.__name__)

    # class or class instance
    if hasattr(func, '__call__'):
        # class
        if hasattr(func, '__name__'):
            return func.__name__

        # instance of a class with a __call__ method
        return func.__class__.__name__

    raise TypeError('Unable to determine a name for %r -- maybe it is not a callable?' % func)


def obj_to_ref(obj):
    """
    将函数或方法等callable对象，序列化成可直接引用的有效包与路径引用串
    Returns the path to the given callable.

    :rtype: str
    :raises TypeError: if the given object is not callable
    :raises ValueError: if the given object is a :class:`~functools.partial`, lambda or a nested
        function

    """
    if isinstance(obj, partial):
        raise ValueError('Cannot create a reference to a partial()')

    name = get_callable_name(obj)
    if '<lambda>' in name:
        raise ValueError('Cannot create a reference to a lambda')
    if '<locals>' in name:
        raise ValueError('Cannot create a reference to a nested function')

    if ismethod(obj):
        if hasattr(obj, 'im_self') and obj.im_self:
            # bound method
            module = obj.im_self.__module__
        elif hasattr(obj, 'im_class') and obj.im_class:
            # unbound method
            module = obj.im_class.__module__
        else:
            module = obj.__module__
    else:
        module = obj.__module__
    return '%s:%s' % (module, name)


def ref_to_obj(ref):
    """
    将包与路径引用串反序列化成可执行对象
    Returns the object pointed to by ``ref``.
    :type ref: str
    """
    if not isinstance(ref, six.string_types):
        raise TypeError('References must be strings')
    if ':' not in ref:
        raise ValueError('Invalid reference')

    module_name, rest = ref.split(':', 1)
    try:
        obj = __import__(module_name, fromlist=[rest])
    except ImportError:
        raise LookupError('Error resolving reference %s: could not import module' % ref)

    try:
        for name in rest.split('.'):
            obj = getattr(obj, name)
        return obj
    except Exception:
        raise LookupError('Error resolving reference %s: error looking up object' % ref)
